package org.imageclef.flickr;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.TreeSet;

public class Evaluations2012 {

	public final static int IMAGECOUNT = 10000;
	public final static int CONCEPTCOUNT = 94;
	public final static int QUERYCOUNT = 42;
	public final static int QUERYTOP = 100;
	public final static int APLEVELS = 11;
	public final static double EPSILON = 0.0000000001;

	// prepare decimal formatter
	private static DecimalFormat formatter = new DecimalFormat("0.0000");

	// enum for indicating true/false positives/negatives
	private static enum TPFPTNFN {
		TP,
		FP,
		TN,
		FN
	}

	public static boolean annotationBinary(String groundtruth, String input, String output) {
		// list all files in the groundtruth directory
		File[] gts = (new File(groundtruth)).listFiles(new ExtensionFilter(".txt"));
		if (gts == null || gts.length == 0) {
			System.err.println("ERROR: no files found in " + groundtruth);
			return false;
		}
		// list all files in the input directory
		File[] runs = (new File(input)).listFiles(new ExtensionFilter(".txt"));
		if (runs == null || runs.length == 0) {
			System.err.println("ERROR: no files found in " + input);
			return false;
		}
		// create output directory
		File store = new File(output);
		if (!store.mkdir() && !store.exists()) {
			System.err.println("ERROR: could not create output directory for " + output);
			return false;
		}
		// load ground truth
		// note: the ground truth contains for each concept which images it was considered to
		//       be present in, as determined by majority vote of the crowdflower workers
		// note: determine the various concept categories
		TreeMap<Integer, KeyValuePair<String, HashSet<String>>> annotations = new TreeMap<Integer, KeyValuePair<String, HashSet<String>>>();
		TreeMap<String, TreeSet<Integer>> categories = new TreeMap<String, TreeSet<Integer>>();
		for (File gt : gts) {
			// split the file name into concept identifier and concept name
			String[] split = Extension.remove(gt.getName()).split("\\s");
			if (split.length != 2) {
				System.err.println("ERROR: invalid concept file " + gt.getPath());
				return false;
			}
			int id;
			try {
				id = Integer.parseInt(split[0]);
				if (id < 0 || id > CONCEPTCOUNT-1) {
					System.err.println("ERROR: invalid data in " + gt.getPath());
					return false;
				}
			}
			catch (NumberFormatException e) {
				System.err.println("ERROR: could not parse concept filename " + gt.getPath());
				return false;
			}
			String name = split[1];
			// determine the concept category
			String[] cname = name.split("_");
			if (cname.length != 2) {
				System.err.println("ERROR: invalid data in " + gt.getPath());
				return false;
			}
			TreeSet<Integer> cid = categories.get("_" + cname[0]);
			if (cid == null) {
				cid = new TreeSet<Integer>();
				categories.put("_" + cname[0], cid);
			}
			cid.add(id);
			// also add concept identifier to the all category
			cid = categories.get("");
			if (cid == null) {
				cid = new TreeSet<Integer>();
				categories.put("", cid);
			}
			cid.add(id);
			// load the image identifiers associated with this concept
			LinkedList<String> ids = FileReader.ReadTextLines(gt);
			if (ids == null || ids.size() == 0) {
				System.err.println("ERROR: invalid data in concept file " + gt.getPath());
				return false;
			}
			// add to annotations
			annotations.put(id, new KeyValuePair<String, HashSet<String>>(name, new HashSet<String>(ids)));
		}
		// walk through the input files
		for (File run : runs) {
			// prepare arrays
			TPFPTNFN[][] tpfptnfn = new TPFPTNFN[IMAGECOUNT][CONCEPTCOUNT];
			String[] imagemap = new String[IMAGECOUNT];
			// process input file
			BufferedReader reader;
			try { reader = new BufferedReader(new InputStreamReader(new FileInputStream(run))); }
			catch (IOException e) {
				System.err.println("ERROR: could not open " + run.getPath());
				return false;
			}
			try {
				// verify the file refers to an annotation run
				System.out.println("verifying input file  " + run.getPath());
				String s = reader.readLine();
				if (s == null) {
					System.err.println("ERROR: no data in " + run.getPath());
					continue;
				}
				String[] check = s.split("\\s");
				if (check.length != 1+CONCEPTCOUNT*2) {
					System.err.println("ERROR: invalid file format in " + run.getPath());
					continue;
				}
				// process the run
				System.out.println("processing input file " + run.getPath());
				for (int i = 0; ; i++) {
					if (i > IMAGECOUNT-1) {
						System.err.println("ERROR: invalid data in " + run.getPath());
						return false;
					}
					// parse the line
					String[] split = s.split("\\s");
					// extract image identifier
					String id = Extension.remove(split[0]);
					// map image identifier
					imagemap[i] = id;
					// walk through the binary annotations
					for (int j = 0; j < CONCEPTCOUNT*2; j+=2) {
						int c = j / 2;
						String binary = split[j+2];
						// compare annotation with ground truth
						if (annotations.get(c).value.contains(id)) {
							if (binary.equals("0"))
								tpfptnfn[i][c] = TPFPTNFN.FN;
							else if (binary.equals("1"))
								tpfptnfn[i][c] = TPFPTNFN.TP;
							else {
								System.err.println("ERROR: invalid data in concept file " + run.getPath());
								return false;
							}
						}
						else {
							if (binary.equals("0"))
								tpfptnfn[i][c] = TPFPTNFN.TN;
							else if (binary.equals("1"))
								tpfptnfn[i][c] = TPFPTNFN.FP;
							else {
								System.err.println("ERROR: invalid data in concept file " + run.getPath());
								return false;
							}
						}
					}
					// read the next line from the file
					s = reader.readLine();
					if (s == null)
						break;
				}
			} catch (IOException e) {
				System.err.println("ERROR: could not read from " + run.getPath());
				System.err.println(e.getMessage()); e.printStackTrace();
				return false;
			}
			finally {
				try { reader.close(); }
				catch (IOException e) {
					System.err.println("ERROR: could not close " + run.getPath());
					System.err.println(e.getMessage()); e.printStackTrace();
					return false;
				}
			}
			// process the annotations for the different concept categories on a per image basis
			for (Entry<String, TreeSet<Integer>> entry : categories.entrySet()) {
				// extract the name of the institute that submitted this run
				LinkedList<String> lines = FileReader.ReadTextLines(new File(run.getParent().toString(), Extension.remove(run.getName()) + ".xml"));
				String owner = "";
				for (String line : lines) {
					if (line.startsWith("<owner>") && line.endsWith("</owner>")) {
						owner = line.substring("<owner>".length(), line.length() - "</owner>".length());
						break;
					}
				}
				File file = new File(store, owner);
				if (!file.mkdir() && !file.exists()) {
					System.err.println("ERROR: could not create output directory for " + file);
					return false;
				}
				file = new File(file, Extension.remove(run.getName()));
				if (!file.mkdir() && !file.exists()) {
					System.err.println("ERROR: could not create output directory for " + file);
					return false;
				}
				file = new File(file, "image" + entry.getKey() + ".txt");
				// write the results to disk
				System.out.println("saving output file    " + file.getPath());
				BufferedWriter writer;
				try { writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file))); }
				catch (IOException e) {
					System.err.println("ERROR: could not open " + file.getName());
					System.err.println(e.getMessage()); e.printStackTrace();					
					return false;
				}
				try {
					// measure besides individual image accuracy also the micro and macro accuracies for all images combined
					Accuracy overall = new Accuracy();
					// write header
					writer.write("image");
					writer.write('\t');
					writer.write("true_positive");
					writer.write(' ');
					writer.write("false_positive");
					writer.write(' ');
					writer.write("true_negative");
					writer.write(' ');
					writer.write("false_negative");
					writer.write(' ');
					writer.write("precision");
					writer.write(' ');
					writer.write("recall");
					writer.write(' ');
					writer.write("f1");
					writer.newLine();
					// walk through the images
					for (int i = 0; i < IMAGECOUNT; i++) {
						Accuracy individual = new Accuracy();
						// walk through the concepts
						for (int c = 0; c < CONCEPTCOUNT; c++) {
							// only include this concept if it is on the list
							if (!entry.getValue().contains(c))
								continue;
							// count the true/false positive/negative
							switch (tpfptnfn[i][c]) {
							case TP: individual.tp++; overall.tp++; break;
							case FP: individual.fp++; overall.fp++; break;
							case TN: individual.tn++; overall.tn++; break;
							case FN: individual.fn++; overall.fn++; break;
							}
						}
						// when we are evaluating a particular category it is not always the case
						// that one of the concepts belonging to this category are present in this
						// image. therefore we introduce the 'no-concept' concept, which activates
						// when no concepts were indicated to be present in the image.
						// note: the extra concept does not influence the precision and recall values
						//       for images where at least one concept is present, since it will be
						//       treated as a true negative, which is a value that is not used in the
						//       calculation of precision and recall.
						if (individual.tp == 0 && individual.fp == 0 && individual.fn == 0 && individual.tn != 0) {
							individual.tp++;
							overall.tp++;
						}
						else {
							individual.tn++;
							overall.tn++;
						}
						// determine individual precision, recall and f1-measure
						individual.precision();
						individual.recall();
						individual.f1();
						overall.p += individual.p;
						overall.r += individual.r;
						overall.f += individual.f;
						// write to disk
						writer.write(imagemap[i]);
						writer.write('\t');
						writer.write(Integer.toString(individual.tp));
						writer.write(' ');
						writer.write(Integer.toString(individual.fp));
						writer.write(' ');
						writer.write(Integer.toString(individual.tn));
						writer.write(' ');
						writer.write(Integer.toString(individual.fn));
						writer.write(' ');
						writer.write(formatter.format(individual.p));
						writer.write(' ');
						writer.write(formatter.format(individual.r));
						writer.write(' ');
						writer.write(formatter.format(individual.f));
						writer.newLine();
					}
					// write instance average to disk
					// note: instance average is determined by averaging the individual precisions,
					//       recalls and f1-scores
					// note: we have already stored the cumulative individual precisions, recalls
					//       and f1-scores, and therefore only have to average them
					overall.p /= IMAGECOUNT;
					overall.r /= IMAGECOUNT;
					overall.f /= IMAGECOUNT;
					writer.write("instance");
					writer.write('\t');
					writer.write(Integer.toString(overall.tp));
					writer.write(' ');
					writer.write(Integer.toString(overall.fp));
					writer.write(' ');
					writer.write(Integer.toString(overall.tn));
					writer.write(' ');
					writer.write(Integer.toString(overall.fn));
					writer.write(' ');
					writer.write(formatter.format(overall.p));
					writer.write(' ');
					writer.write(formatter.format(overall.r));
					writer.write(' ');
					writer.write(formatter.format(overall.f));
					writer.newLine();
					// write macro average to disk
					// note: macro average is determined by first averaging the individual precisions
					//       and recalls, and then computing the f1-measure from these averages. since
					//       we already averaged the precisions and recalls all we need to do now is
					//       recompute the f1-measure.
					overall.f1();
					writer.write("macro");
					writer.write('\t');
					writer.write(Integer.toString(overall.tp));
					writer.write(' ');
					writer.write(Integer.toString(overall.fp));
					writer.write(' ');
					writer.write(Integer.toString(overall.tn));
					writer.write(' ');
					writer.write(Integer.toString(overall.fn));
					writer.write(' ');
					writer.write(formatter.format(overall.p));
					writer.write(' ');
					writer.write(formatter.format(overall.r));
					writer.write(' ');
					writer.write(formatter.format(overall.f));
					writer.newLine();
					// write micro average to disk
					// note: micro average is determined by first summing the tp, fp, tn and fn values,
					//       and then determining the precision, recall and f1-measure from them.
					// note: we have already stored the cumulative tp, fp, tn and fn values, and therefore
					//       only have to recompute the precision, recall and f1-measure.
					overall.precision();
					overall.recall();
					overall.f1();
					writer.write("micro");
					writer.write('\t');
					writer.write(Integer.toString(overall.tp));
					writer.write(' ');
					writer.write(Integer.toString(overall.fp));
					writer.write(' ');
					writer.write(Integer.toString(overall.tn));
					writer.write(' ');
					writer.write(Integer.toString(overall.fn));
					writer.write(' ');
					writer.write(formatter.format(overall.p));
					writer.write(' ');
					writer.write(formatter.format(overall.r));
					writer.write(' ');
					writer.write(formatter.format(overall.f));
				} catch (IOException e) {
					System.err.println("ERROR: could not write to " + file.getName());
					System.err.println(e.getMessage()); e.printStackTrace();
					return false;
				} finally {
					try { writer.close(); }
					catch (IOException e) {
						System.err.println("ERROR: could not close " + file.getName());
						System.err.println(e.getMessage()); e.printStackTrace();
					}
				}
			}
			// process the annotations on a per concept category basis
			for (Entry<String, TreeSet<Integer>> entry : categories.entrySet()) {
				// extract the name of the institute that submitted this run
				LinkedList<String> lines = FileReader.ReadTextLines(new File(run.getParent().toString(), Extension.remove(run.getName()) + ".xml"));
				String owner = "";
				for (String line : lines) {
					if (line.startsWith("<owner>") && line.endsWith("</owner>")) {
						owner = line.substring("<owner>".length(), line.length() - "</owner>".length());
						break;
					}
				}
				File file = new File(store, owner);
				if (!file.mkdir() && !file.exists()) {
					System.err.println("ERROR: could not create output directory for " + file);
					return false;
				}
				file = new File(file, Extension.remove(run.getName()));
				if (!file.mkdir() && !file.exists()) {
					System.err.println("ERROR: could not create output directory for " + file);
					return false;
				}
				file = new File(file, "concept" + entry.getKey() + ".txt");
				// write the results to disk
				System.out.println("saving output file    " + file.getPath());
				BufferedWriter writer;
				try { writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file))); }
				catch (IOException e) {
					System.err.println("ERROR: could not open " + file.getName());
					System.err.println(e.getMessage()); e.printStackTrace();					
					return false;
				}
				try {
					// measure besides individual concept accuracy also the micro and macro accuracies for all concepts combined
					Accuracy overall = new Accuracy();
					// write header
					writer.write("concept");
					writer.write('\t');
					writer.write("true_positive");
					writer.write(' ');
					writer.write("false_positive");
					writer.write(' ');
					writer.write("true_negative");
					writer.write(' ');
					writer.write("false_negative");
					writer.write(' ');
					writer.write("precision");
					writer.write(' ');
					writer.write("recall");
					writer.write(' ');
					writer.write("f1");
					writer.newLine();
					// walk through the images
					for (int c = 0; c < CONCEPTCOUNT; c++) {
						// only include this concept if it is on the list
						if (!entry.getValue().contains(c))
							continue;
						// walk through the images
						Accuracy individual = new Accuracy();
						for (int i = 0; i < IMAGECOUNT; i++) {
							// count the true/false positive/negative
							switch (tpfptnfn[i][c]) {
							case TP: individual.tp++; overall.tp++; break;
							case FP: individual.fp++; overall.fp++; break;
							case TN: individual.tn++; overall.tn++; break;
							case FN: individual.fn++; overall.fn++; break;
							}
						}
						// determine individual precision, recall and f1-measure
						individual.precision();
						individual.recall();
						individual.f1();
						overall.p += individual.p;
						overall.r += individual.r;
						overall.f += individual.f;
						// write to disk
						writer.write(Integer.toString(c) + " " + annotations.get(c).key);
						writer.write('\t');
						writer.write(Integer.toString(individual.tp));
						writer.write(' ');
						writer.write(Integer.toString(individual.fp));
						writer.write(' ');
						writer.write(Integer.toString(individual.tn));
						writer.write(' ');
						writer.write(Integer.toString(individual.fn));
						writer.write(' ');
						writer.write(formatter.format(individual.p));
						writer.write(' ');
						writer.write(formatter.format(individual.r));
						writer.write(' ');
						writer.write(formatter.format(individual.f));
						writer.newLine();
					}
					// write instance average to disk
					// note: make sure to divide by the number of concepts in the category
					//       and not the total number of concepts
					overall.p /= entry.getValue().size();
					overall.r /= entry.getValue().size();
					overall.f /= entry.getValue().size();
					writer.write("instance");
					writer.write('\t');
					writer.write(Integer.toString(overall.tp));
					writer.write(' ');
					writer.write(Integer.toString(overall.fp));
					writer.write(' ');
					writer.write(Integer.toString(overall.tn));
					writer.write(' ');
					writer.write(Integer.toString(overall.fn));
					writer.write(' ');
					writer.write(formatter.format(overall.p));
					writer.write(' ');
					writer.write(formatter.format(overall.r));
					writer.write(' ');
					writer.write(formatter.format(overall.f));
					writer.newLine();
					// write macro average to disk
					overall.f1();
					writer.write("macro");
					writer.write('\t');
					writer.write(Integer.toString(overall.tp));
					writer.write(' ');
					writer.write(Integer.toString(overall.fp));
					writer.write(' ');
					writer.write(Integer.toString(overall.tn));
					writer.write(' ');
					writer.write(Integer.toString(overall.fn));
					writer.write(' ');
					writer.write(formatter.format(overall.p));
					writer.write(' ');
					writer.write(formatter.format(overall.r));
					writer.write(' ');
					writer.write(formatter.format(overall.f));
					writer.newLine();
					// write micro average to disk
					overall.precision();
					overall.recall();
					overall.f1();
					writer.write("micro");
					writer.write('\t');
					writer.write(Integer.toString(overall.tp));
					writer.write(' ');
					writer.write(Integer.toString(overall.fp));
					writer.write(' ');
					writer.write(Integer.toString(overall.tn));
					writer.write(' ');
					writer.write(Integer.toString(overall.fn));
					writer.write(' ');
					writer.write(formatter.format(overall.p));
					writer.write(' ');
					writer.write(formatter.format(overall.r));
					writer.write(' ');
					writer.write(formatter.format(overall.f));
				} catch (IOException e) {
					System.err.println("ERROR: could not write to " + file.getName());
					System.err.println(e.getMessage()); e.printStackTrace();
					return false;
				} finally {
					try { writer.close(); }
					catch (IOException e) {
						System.err.println("ERROR: could not close " + file.getName());
						System.err.println(e.getMessage()); e.printStackTrace();
					}
				}
			}
		}
		return true;
	}

	public static boolean annotationConfidence(String groundtruth, String input, String output) {
		// list all files in the groundtruth directory
		File[] gts = (new File(groundtruth)).listFiles(new ExtensionFilter(".txt"));
		if (gts == null || gts.length == 0) {
			System.err.println("ERROR: no files found in " + groundtruth);
			return false;
		}
		// list all files in the input directory
		File[] runs = (new File(input)).listFiles(new ExtensionFilter(".txt"));
		if (runs == null || runs.length == 0) {
			System.err.println("ERROR: no files found in " + input);
			return false;
		}
		// create output directory
		File store = new File(output);
		if (!store.mkdir() && !store.exists()) {
			System.err.println("ERROR: could not create output directory for " + output);
			return false;
		}
		// load ground truth
		// note: the ground truth contains for each concept which images it was considered to
		//       be present in, as determined by majority vote of the crowdflower workers
		// note: determine the various concept categories
		TreeMap<Integer, KeyValuePair<String, HashSet<String>>> annotations = new TreeMap<Integer, KeyValuePair<String, HashSet<String>>>();
		TreeMap<String, TreeSet<Integer>> categories = new TreeMap<String, TreeSet<Integer>>();
		for (File gt : gts) {
			// split the file name into concept identifier and concept name
			String[] split = Extension.remove(gt.getName()).split("\\s");
			if (split.length != 2) {
				System.err.println("ERROR: invalid concept file " + gt.getPath());
				return false;
			}
			int id;
			try {
				id = Integer.parseInt(split[0]);
				if (id < 0 || id > CONCEPTCOUNT-1) {
					System.err.println("ERROR: invalid data in " + gt.getPath());
					return false;
				}
			}
			catch (NumberFormatException e) {
				System.err.println("ERROR: could not parse concept filename " + gt.getPath());
				return false;
			}
			String name = split[1];
			// determine the concept category
			String[] cname = name.split("_");
			if (cname.length != 2) {
				System.err.println("ERROR: invalid data in " + gt.getPath());
				return false;
			}
			TreeSet<Integer> cid = categories.get("_" + cname[0]);
			if (cid == null) {
				cid = new TreeSet<Integer>();
				categories.put("_" + cname[0], cid);
			}
			cid.add(id);
			// also add concept identifier to the all category
			cid = categories.get("");
			if (cid == null) {
				cid = new TreeSet<Integer>();
				categories.put("", cid);
			}
			cid.add(id);
			// load the image identifiers associated with this concept
			LinkedList<String> ids = FileReader.ReadTextLines(gt);
			if (ids == null || ids.size() == 0) {
				System.err.println("ERROR: invalid data in concept file " + gt.getPath());
				return false;
			}
			// add to annotations
			annotations.put(id, new KeyValuePair<String, HashSet<String>>(name, new HashSet<String>(ids)));
		}
		// walk through the input files
		for (File run : runs) {
			// prepare arrays
			double[][] confidences = new double[CONCEPTCOUNT][IMAGECOUNT];
			String[] imagemap = new String[IMAGECOUNT];
			// process input file
			BufferedReader reader;
			try { reader = new BufferedReader(new InputStreamReader(new FileInputStream(run))); }
			catch (IOException e) {
				System.err.println("ERROR: could not open " + run.getPath());
				return false;
			}
			try {
				// verify the file refers to an annotation run
				System.out.println("verifying input file  " + run.getPath());
				String s = reader.readLine();
				if (s == null) {
					System.err.println("ERROR: no data in " + run.getPath());
					continue;
				}
				String[] check = s.split("\\s");
				if (check.length != 1+CONCEPTCOUNT*2) {
					System.err.println("ERROR: invalid file format in " + run.getPath());
					continue;
				}
				// process the run
				System.out.println("processing input file " + run.getPath());
				for (int i = 0; ; i++) {
					if (i > IMAGECOUNT-1) {
						System.err.println("ERROR: invalid data in " + run.getPath());
						return false;
					}
					// parse the line
					String[] split = s.split("\\s");
					// extract image identifier
					String id = Extension.remove(split[0]);
					// map image identifier
					imagemap[i] = id;
					// walk through the confidence annotations
					for (int j = 0; j < CONCEPTCOUNT*2; j+=2) {
						int c = j / 2;
						double confidence;
						try { confidence = Double.parseDouble(split[j+1]); }
						catch (NumberFormatException e) {
							System.err.println("ERROR: invalid data in concept file " + run.getPath());
							return false;
						}
						// add to array
						confidences[c][i] = confidence;
					}
					// read the next line from the file
					s = reader.readLine();
					if (s == null)
						break;
				}
			} catch (IOException e) {
				System.err.println("ERROR: could not read from " + run.getPath());
				System.err.println(e.getMessage()); e.printStackTrace();
				return false;
			}
			finally {
				try { reader.close(); }
				catch (IOException e) {
					System.err.println("ERROR: could not close " + run.getPath());
					System.err.println(e.getMessage()); e.printStackTrace();
					return false;
				}
			}
			// process the confidence scores
			ArrayList<Double> ap_n = new ArrayList<Double>(CONCEPTCOUNT);
			ArrayList<Double> ap_i = new ArrayList<Double>(CONCEPTCOUNT);
			for (int c = 0; c < CONCEPTCOUNT; c++) {
				// sort the images by their confidence, from high to low, for this concept
				ArrayList<KeyValuePair<String, Double>> scores = new ArrayList<KeyValuePair<String,Double>>();
				for (int i = 0; i < IMAGECOUNT; i++)
					scores.add(new KeyValuePair<String, Double>(imagemap[i], confidences[c][i]));
				Collections.sort(scores, new Comparator<KeyValuePair<String, Double>>() {
					public int compare(final KeyValuePair<String, Double> kvp1, final KeyValuePair<String, Double> kvp2) {
						return -kvp1.value.compareTo(kvp2.value);
					}
				});
				// determine tie-aware non-interpolated average precision
				int preRel = 0;
				int i = 0;
				double sum = 0.0;
				LinkedList<KeyValuePair<Double, Double>> pr = new LinkedList<KeyValuePair<Double, Double>>();
				while (i < IMAGECOUNT) {
					int tieRel = 0;
					int tieStart = i;
					for (; i < IMAGECOUNT; i++) {
						KeyValuePair<String, Double> current = scores.get(i); 
						// note: be careful about java's automatic (un)boxing of primitive types. here we have
						//       to explicit compare the values as otherwise the references will be compared
						//       giving unexpected and undesired behavior.
						if (i > tieStart && current.value.doubleValue() != scores.get(i-1).value.doubleValue())
							break;
						if (annotations.get(c).value.contains(current.key))
							tieRel++;
					}
					int tieLen = i - tieStart;
					double f = 1.0 * tieRel / tieLen;
					double g = tieLen == 1 ? 0.0 : (tieRel - 1.0) / (tieLen - 1.0);
					for (int j = tieStart + 1; j <= i; j++)
						sum += (f * (preRel + (j - tieStart - 1) * g + 1.0)) / j;
					preRel += tieRel;
					// add precision at current recall level
					if (tieRel > 0) {
						double p = (double)preRel / (double)annotations.get(c).value.size();
						double r = (double)preRel / (double)i;
						pr.add(new KeyValuePair<Double, Double>(p, r));
					}
				}
				ap_n.add(sum / (double)annotations.get(c).value.size());
				// determine interpolated average precision at recall levels ranging from zero to one in
				// small steps
				// note: the precision at a certain recall level is the maximum precision obtained at any
				//       recall level equal or higher to the one we're looking at
				//       e.g. if we want to determine the precision at recall level 0.2 and at recall level
				//            0.33 the precision happens to be the largest value for recall level 0.2 and up,
				//            then this is the value we report for recall level 0.2
				double a = 0.0;
				for (int z = 0; z < APLEVELS; z++) {
					double l = (double)z / (double)(APLEVELS - 1);
					double pmax = 0.0;
					for (KeyValuePair<Double, Double> e : pr) {
						if (l > e.key)
							continue;
						if (pmax < e.value)
							pmax = e.value;
					}
					a += pmax;
				}
				ap_i.add(a / APLEVELS);
			}
			// process the concept categories
			for (Entry<String, TreeSet<Integer>> entry : categories.entrySet()) {
				// determine map and gmap
				// note: for map we simply sum the average precisions for each concept and then average them,
				//       whereas for gmap we sum the logs of the average precisions for each concept, average
				//       them and then exponentiate this average.
				// note: to deal with average precisions that are zero, we add a very tiny epsilon value that
				//       we afterwards remove again.
				double map_n = 0.0;
				double gmap_n = 0.0;
				double map_i = 0.0;
				double gmap_i = 0.0;
				for (int c = 0; c < CONCEPTCOUNT; c++) {
					// only include this concept if it is on the list
					if (!entry.getValue().contains(c))
						continue;
					// add average precisions
					map_n += ap_n.get(c);
					map_i += ap_i.get(c);
					gmap_n += Math.log(ap_n.get(c) + EPSILON);
					gmap_i += Math.log(ap_i.get(c) + EPSILON);
				}
				map_n /= entry.getValue().size();
				gmap_n /= entry.getValue().size();
				gmap_n = Math.exp(gmap_n) - EPSILON;
				map_i /= entry.getValue().size();
				gmap_i /= entry.getValue().size();
				gmap_i = Math.exp(gmap_i) - EPSILON;
				// extract the name of the institute that submitted this run
				LinkedList<String> lines = FileReader.ReadTextLines(new File(run.getParent().toString(), Extension.remove(run.getName()) + ".xml"));
				String owner = "";
				for (String line : lines) {
					if (line.startsWith("<owner>") && line.endsWith("</owner>")) {
						owner = line.substring("<owner>".length(), line.length() - "</owner>".length());
						break;
					}
				}
				File file = new File(store, owner);
				if (!file.mkdir() && !file.exists()) {
					System.err.println("ERROR: could not create output directory for " + file);
					return false;
				}
				file = new File(file, Extension.remove(run.getName()));
				if (!file.mkdir() && !file.exists()) {
					System.err.println("ERROR: could not create output directory for " + file);
					return false;
				}
				file = new File(file, "ap" + entry.getKey() + ".txt");
				// write the results to disk
				System.out.println("saving output file    " + file.getPath());
				BufferedWriter writer;
				try { writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file))); }
				catch (IOException e) {
					System.err.println("ERROR: could not open " + file.getName());
					System.err.println(e.getMessage()); e.printStackTrace();					
					return false;
				}
				try {
					writer.write("concept");
					writer.write('\t');
					writer.write("ap_n");
					writer.write(' ');
					writer.write("ap_i");
					writer.newLine();
					// walk through the concepts
					for (int c = 0; c < CONCEPTCOUNT; c++) {
						if (!entry.getValue().contains(c))
							continue;
						writer.write(Integer.toString(c) + " " + annotations.get(c).key);
						writer.write('\t');
						writer.write(formatter.format(ap_n.get(c)));
						writer.write(' ');
						writer.write(formatter.format(ap_i.get(c)));
						writer.newLine();
					}
					// write map
					writer.write("map_n");
					writer.write('\t');
					writer.write(formatter.format(map_n));
					writer.newLine();
					writer.write("map_i");
					writer.write('\t');
					writer.write(formatter.format(map_i));
					writer.newLine();
					// write gmap
					writer.write("gmap_n");
					writer.write('\t');
					writer.write(formatter.format(gmap_n));
					writer.newLine();
					writer.write("gmap_i");
					writer.write('\t');
					writer.write(formatter.format(gmap_i));
				} catch (IOException e) {
					System.err.println("ERROR: could not write to " + file.getName());
					System.err.println(e.getMessage()); e.printStackTrace();
					return false;
				} finally {
					try { writer.close(); }
					catch (IOException e) {
						System.err.println("ERROR: could not close " + file.getName());
						System.err.println(e.getMessage()); e.printStackTrace();
					}
				} 
			}
		}
		return true;
	}

	public static boolean retrievalEvaluation(String groundtruth, String input, String output) {
		// list all files in the groundtruth directory
		File[] gts = (new File(groundtruth)).listFiles(new ExtensionFilter(".txt"));
		if (gts == null || gts.length == 0) {
			System.err.println("ERROR: no files found in " + groundtruth);
			return false;
		}
		// list all files in the input directory
		File[] runs = (new File(input)).listFiles(new ExtensionFilter(".txt"));
		if (runs == null || runs.length == 0) {
			System.err.println("ERROR: no files found in " + input);
			return false;
		}
		// create output directory
		File store = new File(output);
		if (!store.mkdir() && !store.exists()) {
			System.err.println("ERROR: could not create output directory for " + output);
			return false;
		}
		// load ground truth
		// note: the ground truth contains for each query which images were considered to
		//       be relevant, as determined by majority vote of the crowdflower workers
		TreeMap<Integer, HashSet<String>> annotations = new TreeMap<Integer, HashSet<String>>();
		for (File gt : gts) {
			// convert file name to query identifier
			int id;
			try {
				id = Integer.parseInt(Extension.remove(gt.getName()));
			} catch (NumberFormatException e) {
				System.err.println("ERROR: could not parse query identifier from filename " + gt.getPath());
				return false;
			}
			// load the image identifiers associated with this concept
			LinkedList<String> ids = FileReader.ReadTextLines(gt);
			if (ids == null || ids.size() == 0) {
				System.err.println("ERROR: invalid data in concept filen " + gt.getPath());
				return false;
			}
			// add to annotations
			annotations.put(id, new HashSet<String>(ids));
		}
		// walk through each run
		for (File run : runs) {
			System.out.println("processing run " + run.getPath());
			// prepare arrays
			double[][] confidences = new double[QUERYCOUNT][QUERYTOP];
			for (int i = 0; i < confidences.length; i++)
				Arrays.fill(confidences[i], Double.NaN);
			String[][] imagemap = new String[QUERYCOUNT][QUERYTOP];
			// process input file
			BufferedReader reader;
			try { reader = new BufferedReader(new InputStreamReader(new FileInputStream(run))); }
			catch (IOException e) {
				System.err.println("ERROR: could not open " + run.getPath());
				return false;
			}
			try {
				// verify the file refers to a retrieval run
				String s = reader.readLine();
				if (s == null) {
					System.err.println("ERROR: no data in " + run.getPath());
					continue;
				}
				String[] check = s.split("\\s");
				if (check.length != 6 || !check[1].equals("IC12")) {
					System.err.println("ERROR: invalid file format in " + run.getPath());
					continue;
				}
				// walk through the images for the specified query
				do {
					String[] split = s.split("\\s");
					// extract the rank
					int r = Integer.parseInt(split[3]);
					if (r >= 0 && r < QUERYTOP) {
						// check the query number
						int query = Integer.parseInt(split[0]);
						// extract the image identifier, without the extension
						String id = Extension.remove(split[2]);
						// map image identifier
						imagemap[query][r] = id;
						// extract the confidence
						double confidence = Double.parseDouble(split[4]);
						// add to array
						confidences[query][r] = confidence;
					}
					// read the next line from the file
					s = reader.readLine();
					if (s == null)
						break;
				} while (true);
			} catch (IOException e) {
				System.err.println("ERROR: could not read from " + run.getPath());
				System.err.println(e.getMessage()); e.printStackTrace();
				return false;
			}
			finally {
				try { reader.close(); }
				catch (IOException e) {
					System.err.println("ERROR: could not close " + run.getPath());
					System.err.println(e.getMessage()); e.printStackTrace();
					return false;
				}
			}
			// process the confidence scores
			double[] ap_n = new double[QUERYCOUNT];
			double[] ap_i = new double[QUERYCOUNT];
			double[][] p_x = new double[10][QUERYCOUNT];
			for (int q = 0; q < QUERYCOUNT; q++) {
				// sort the images by their confidence, from high to low, for this query
				ArrayList<KeyValuePair<String, Double>> scores = new ArrayList<KeyValuePair<String,Double>>();
				for (int i = 0; i < QUERYTOP; i++) {
					// check whether each confidence element was filled in
					// note: in principle we should be able to handle runs with less images returned,
					//       but knowing that each run has at least the required number of images makes
					//       processing the data a lot easier
					if (Double.isNaN(confidences[q][i])) {
						System.err.println("ERROR: insufficient images for query " + q + " in " + run.getPath());
						return false;
					}
					// add to array
					scores.add(new KeyValuePair<String, Double>(imagemap[q][i], confidences[q][i]));
				}
				Collections.sort(scores, new Comparator<KeyValuePair<String, Double>>() {
					public int compare(final KeyValuePair<String, Double> kvp1, final KeyValuePair<String, Double> kvp2) {
						return -kvp1.value.compareTo(kvp2.value);
					}
				});
				// determine tie-aware non-interpolated average precision
				int preRel = 0;
				int i = 0;
				double sum = 0.0;
				LinkedList<KeyValuePair<Double, Double>> pr = new LinkedList<KeyValuePair<Double, Double>>();
				while (i < QUERYTOP) {
					int tieRel = 0;
					int tieStart = i;
					for (; i < QUERYTOP; i++) {
						KeyValuePair<String, Double> current = scores.get(i); 
						// note: be careful about java's automatic (un)boxing of primitive types. here we have
						//       to explicit compare the values as otherwise the references will be compared
						//       giving unexpected and undesired behavior.
						if (i > tieStart && current.value.doubleValue() != scores.get(i-1).value.doubleValue())
							break;
						if (annotations.get(q).contains(current.key))
							tieRel++;
					}
					int tieLen = i - tieStart;
					double f = 1.0 * tieRel / tieLen;
					double g = tieLen == 1 ? 0.0 : (tieRel - 1.0) / (tieLen - 1.0);
					for (int j = tieStart + 1; j <= i; j++)
						sum += (f * (preRel + (j - tieStart - 1) * g + 1.0)) / j;
					preRel += tieRel;
					// add precision at current recall level
					double p = (double)preRel / (double)annotations.get(q).size();
					for (int x = 0; x < p_x.length; x++) {
						if (i == (x+1)*10)
							p_x[x][q] = p;
					}
					if (tieRel > 0) {
						double r = (double)preRel / (double)i;
						pr.add(new KeyValuePair<Double, Double>(p, r));
					}
				}
				ap_n[q] = sum / (double)annotations.get(q).size();
				for (int x = 0; x < p_x.length; x++) {
					if (Double.isNaN(p_x[x][q]))
						p_x[x][q] = 0.0;
				}
				// determine interpolated average precision at recall levels ranging from zero to one in
				// small steps
				// note: the precision at a certain recall level is the maximum precision obtained at any
				//       recall level equal or higher to the one we're looking at
				//       e.g. if we want to determine the precision at recall level 0.2 and at recall level
				//            0.33 the precision happens to be the largest value for recall level 0.2 and up,
				//            then this is the value we report for recall level 0.2
				double a = 0.0;
				for (int z = 0; z < APLEVELS; z++) {
					double l = (double)z / (double)(APLEVELS - 1);
					double pmax = 0.0;
					for (KeyValuePair<Double, Double> e : pr) {
						if (l > e.key)
							continue;
						if (pmax < e.value)
							pmax = e.value;
					}
					a += pmax;
				}
				ap_i[q] = a / APLEVELS;
			}
			// determine ap@r, map and gmap
			// note: for map we simply sum the average precisions for each query and then average them,
			//       whereas for gmap we sum the logs of the average precisions for each query, average
			//       them and then exponentiate this average.
			// note: to deal with average precisions that are zero, we add a very tiny epsilon value that
			//       we afterwards remove again.
			double map_n = 0.0;
			double gmap_n = 0.0;
			double map_i = 0.0;
			double gmap_i = 0.0;
			double[] ap_r = new double[p_x.length];
			for (int q = 0; q < QUERYCOUNT; q++) {
				map_n += ap_n[q];
				map_i += ap_i[q];
				gmap_n += Math.log(ap_n[q] + EPSILON);
				gmap_i += Math.log(ap_i[q] + EPSILON);
				for (int x = 0; x < p_x.length; x++)
					ap_r[x] += p_x[x][q];
			}
			map_n /= QUERYCOUNT;
			gmap_n /= QUERYCOUNT;
			gmap_n = Math.exp(gmap_n) - EPSILON;
			map_i /= QUERYCOUNT;
			gmap_i /= QUERYCOUNT;
			gmap_i = Math.exp(gmap_i) - EPSILON;
			for (int x = 0; x < p_x.length; x++)
				ap_r[x] /= QUERYCOUNT;
			// extract the name of the institute that submitted this run
			LinkedList<String> lines = FileReader.ReadTextLines(new File(run.getParent().toString(), Extension.remove(run.getName()) + ".xml"));
			String owner = "";
			for (String line : lines) {
				if (line.startsWith("<owner>") && line.endsWith("</owner>")) {
					owner = line.substring("<owner>".length(), line.length() - "</owner>".length());
					break;
				}
			}
			File file = new File(store, owner);
			if (!file.mkdir() && !file.exists()) {
				System.err.println("ERROR: could not create output directory for " + file);
				return false;
			}
			file = new File(file, Extension.remove(run.getName()));
			if (!file.mkdir() && !file.exists()) {
				System.err.println("ERROR: could not create output directory for " + file);
				return false;
			}
			file = new File(file, "ev.txt");
			// write the results to disk		
			System.out.println("saving output file    " + file.getPath());
			BufferedWriter writer;
			try { writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file))); }
			catch (IOException e) {
				System.err.println("ERROR: could not open " + file.getName());
				System.err.println(e.getMessage()); e.printStackTrace();					
				return false;
			}
			try {
				writer.write("query");
				writer.write('\t');
				writer.write("ap_n");
				writer.write(' ');
				writer.write("ap_i");
				for (int x = 0; x < p_x.length; x++) {
					writer.write(' ');
					writer.write("p@" + Integer.toString((x+1)*10));
				}
				writer.newLine();
				// walk through the queries
				for (int q = 0; q < QUERYCOUNT; q++) {
					writer.write(Integer.toString(q));
					writer.write('\t');
					writer.write(formatter.format(ap_n[q]));
					writer.write(' ');
					writer.write(formatter.format(ap_i[q]));
					for (int x = 0; x < p_x.length; x++) {
						writer.write(' ');
						writer.write(formatter.format(p_x[x][q]));
					}
					writer.newLine();
				}
				// write map
				writer.write("map_n");
				writer.write('\t');
				writer.write(formatter.format(map_n));
				writer.newLine();
				writer.write("map_i");
				writer.write('\t');
				writer.write(formatter.format(map_i));
				writer.newLine();
				// write gmap
				writer.write("gmap_n");
				writer.write('\t');
				writer.write(formatter.format(gmap_n));
				writer.newLine();
				writer.write("gmap_i");
				writer.write('\t');
				writer.write(formatter.format(gmap_i));
				// write ap@r
				for (int x = 0; x < p_x.length; x++) {
					writer.newLine();
					writer.write("ap@" + Integer.toString((x+1)*10));
					writer.write('\t');
					writer.write(formatter.format(ap_r[x]));
				}
			} catch (IOException e) {
				System.err.println("ERROR: could not write to " + file.getName());
				System.err.println(e.getMessage()); e.printStackTrace();
				return false;
			} finally {
				try { writer.close(); }
				catch (IOException e) {
					System.err.println("ERROR: could not close " + file.getName());
					System.err.println(e.getMessage()); e.printStackTrace();
				}
			}
		}
		return true;
	}

	public static void main(String[] args) throws Exception {

		/*// evaluate the binary scores of the annotation subtask
		Evaluations2012.annotationBinary("/path/to/test/concept/groundtruth/",
				"/path/to/directory/containing/runs",
				"/path/to/directory/where/to/store/output");

		// evaluate the confidence scores of the annotation subtask
		Evaluations2012.annotationConfidence("/path/to/test/concept/groundtruth/",
				"/path/to/directory/containing/runs",
				"/path/to/directory/where/to/store/output");

		// evaluate the queries of the retrieval subtask
		Evaluations2012.retrievalEvaluation("/path/to/retrieval/querypool/groundtruth/",
				"/path/to/directory/containing/runs",
				"/path/to/directory/where/to/store/output");*/

		System.out.println("done!");
	}

}